package de.hawhamburg.reachability.ui.frame.config.resource;

import java.util.HashMap;
import java.util.Map;

import com.google.gson.Gson;

import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPSubscriber;
import de.hawhamburg.reachability.StopableObject;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.properties.PropertyDescription;
import de.hawhamburg.reachability.properties.json.JsonRAPropertiesResult;
import de.hawhamburg.reachability.ui.frame.config.ReachabilityConfigUI;
import de.hawhamburg.reachability.util.ConnectionUtil;

/**
 * Updates {@link ReachabilityConfigUI} with new configuration sets.<br><br>
 * 
 * This thread shall keep the {@link ReachabilityConfigUI} up to date.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class ReachabilityConfigUIResultListener extends Thread implements StopableObject {
    
    /**
     * The maximum amount of time to wait on response before timeout.  
     */
    private static final long MAX_WAITING   = 3000;
    
    /**
     * The {@link ReachabilityConfigUI} object, which shall be updated.
     */
    private ReachabilityConfigUI internalUI = null;
    
    /**
     * The id of the request message, to match the correct response.
     */
    private String internalRequestId        = null;

    /**
     * Determines if the job has been done or not.
     */
    private boolean jobDone                 = false;

    /**
     * Determines if the method {@link #shutDown()} has been called.
     */
    private boolean shutDown                = false;
    
    /**
     * An {@link LPSubscriber}, which listens on the reachability configuration result topic 
     * (<i>REACHABILITY_CONFIG_RESULT_TOPIC</i>).
     */
    private LPSubscriber resultTopic        = null;
    
    /**
     * The time in milliseconds when this thread has been started.
     */
    private long startingTime;
    
    /**
     * Generates a new {@link ReachabilityConfigUIResultListener} thread to keep the {@link ReachabilityConfigUI} 
     * <i>UI</i> up to date.
     * 
     * @param ui            A {@link ReachabilityConfigUI}, which shall be updated.
     * @param requestId     The ID of the request message to match the correct result. 
     */
    public ReachabilityConfigUIResultListener(final ReachabilityConfigUI ui,
                                              final String requestId) {
        this.internalUI         = ui;
        this.internalRequestId  = requestId;
        this.startingTime       = System.currentTimeMillis();
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        this.shutDown = true;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#isActive()
     * 
     * @return              <code>true</code> if {@link #shutDown()} wasn't called, otherwise <code>false</code>.
     */
    public final boolean isActive() {
        return !this.shutDown;
    }

    /**
     * Connects to the configuration result topic using {@link #init()} and frequently tries to get the newest result 
     * states from this topic. If a new state arrives, the {@link ReachabilityConfigUI} {@link #internalUI} will be
     * updated using {@link de.hawhamburg.reachability.ui.frame.ReachabilityUI#refresh(java.util.Map)}.
     */
    public final void run() {
        init();
        while (isActive() && !timeout()) {
            String msg = ConnectionUtil.getMessageOnTopic(resultTopic);
            if (msg != null) {
                Gson g = new Gson();
                JsonRAPropertiesResult res = g.fromJson(msg, JsonRAPropertiesResult.class);
                if (res.RequestId.equals(this.internalRequestId)) {
                    internalUI.refresh(res.Properties,
                                       fetchPropertiesDescription(res.PropertiesDescription),
                                       res.ServiceStateReference,
                                       this);
                    jobDone = true;
                    shutDown();
                }
            }
        }
        
        // Es gab einen timeout...
        if (!jobDone) {
            internalUI.refresh(new HashMap<String, String>(), null, null, this);
        }
        end();
    }
    
    /**
     * Opens a connection the the configuration result topic (<i>REACHABILITY_CONFIG_RESULT_TOPIC</i>).
     */
    public final void init() {
        String resultTopicName = MainConfig.get("REACHABILITY_CONFIG_RESULT_TOPIC");
        resultTopic = ConnectionUtil.getSubscriber(resultTopicName, this);
    }
    
    /**
     * The method fetches the description of the properties from the result message and groups the properties whith the
     * same <i>display group</i> together. Default group is "<i>Common</i>". 
     * 
     * @param props         The properties with their description as {@link Map} of property name ({@link String}) to 
     *                      {@link PropertyDescription}.
     * @return              A map grouping the properties by their display group as {@link Map} of group 
     *                      ({@link String}) to {@link Map} of property ({@link String}) to {@link PropertyDescription}.
     */
    private Map<String, Map<String, PropertyDescription>>
                fetchPropertiesDescription(final Map<String, PropertyDescription> props) {
        
        Map<String, Map<String, PropertyDescription>> result = new HashMap<String, Map<String, PropertyDescription>>();
        
        for (Map.Entry<String, PropertyDescription> entry : props.entrySet()) {
            String group = "Common";
            if (entry.getValue() != null && entry.getValue().getDisplayGroup() != null) {
                group = entry.getValue().getDisplayGroup();
            }
            
            if (!result.containsKey(group)) {
                result.put(group, new HashMap<String, PropertyDescription>());
            }
            
            result.get(group).put(entry.getKey(), entry.getValue());
        }
        
        return result;
    }
    
    /**
     * Disconnects the {@link #resultTopic}, if it isn't <code>null</code>.
     */
    public final void end() {
        if (resultTopic != null) {
            resultTopic.disconnect();
        }
    }
    
    /**
     * Determines if the waiting-time has run out of acceptance.
     * 
     * @return              <code>true</code> if the service is timed out, otherwise <code>false</code>.
     */
    public final boolean timeout() {
        long timediff = System.currentTimeMillis() - startingTime;
        return timediff > ReachabilityConfigUIResultListener.MAX_WAITING;
    }
}
