package de.hawhamburg.reachability.ui.frame.simulation;

import java.awt.Font;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;

import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.tools.feeder.SensorDataFeeder;
import de.hawhamburg.reachability.ui.UI;

/**
 * Displays the current state of the supplied simulation feeder to the user.<br><br>
 * 
 * This user interfaces displays the current state of the supplied simulation feeder. It gives the user the ability to
 * stop the simulation or to open a more detailed view on the certain events, which has been sent by the simulation
 * feeder.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class ReachabilitySimulationDetailsUI extends UI {
    
    /**
     * Some unique serial version id.
     */
    private static final long serialVersionUID = 5891359533599040221L;
    
    /**
     * The active simulation feeder as {@link SensorDataFeeder}.
     */
    private SensorDataFeeder feeder             = null;
    
    /**
     * Starts and builds a new {@link ReachabilitySimulationDetailsUI}.
     * 
     * @param sourceWindow      The calling panel which will be notified, when the simulation was started as Object 
     *                          implementing {@link JFrame}.
     * @param runningFeeder     The active simulation feeder as {@link SensorDataFeeder}.
     */
    public ReachabilitySimulationDetailsUI(final JFrame sourceWindow,
                                           final SensorDataFeeder runningFeeder) {
        super("Simulation details", sourceWindow);
        this.feeder = runningFeeder;
        build();
    }
    
    /**
     * Starts and builds a new {@link ReachabilitySimulationDetailsUI} by using 
     * {@link #ReachabilitySimulationDetailsUI(JFrame, SensorDataFeeder)} without calling panel.
     * 
     * @param runningFeeder     The active simulation feeder as {@link SensorDataFeeder}.
     */
    public ReachabilitySimulationDetailsUI(final SensorDataFeeder runningFeeder) {
        this(null, runningFeeder);
    }
    
    /**
     * This method adds a menu to the user interface.
     */
    protected final void addMenu() {
        if (this.isStandalone()) {
            JMenu help = new JMenu("Help");
            help.setMnemonic('H');
            
            JMenuItem about = new JMenuItem("About Reachability Agent");
            about.setMnemonic('A');
            about.setName("about");
            help.add(about);
            
            JMenuBar bar = new JMenuBar();
            bar.add(help);
            setJMenuBar(bar);
            
            // Actions
            about.addActionListener(this);
        }
    }
    
    /**
     * Returns a user interface grid.
     * 
     * @param rowsIn            Amount of rows to be displayed
     * @return                  Returns a user interface grid.
     */
    private GridLayout getGrid(final int rowsIn) {
        return getGrid(rowsIn, 2);
    }
    
    /**
     * Adds a new row to the users interface, containing the property and its value.
     * 
     * @param title             The title as {@link String}.
     * @param value             The value as {@link String}.
     */
    private void addInformationRow(final String title,
                                   final String value) {
        JLabel titleLabel = new JLabel(title);
        JLabel valueLabel = new JLabel(value);
        addBody(titleLabel);
        addBody(valueLabel);
    }
    
    /**
     * Adds a new topic row to the user interface.
     * 
     * @param topic             The topics name as {@link String}.
     */
    private void addTopicRow(final String topic) {
        JLabel top      = new JLabel(topic);
        Font parentFont = getBody().getFont();
        top.setFont(new Font(parentFont.getFamily(), Font.BOLD, parentFont.getSize()));
        addBody(top);
        addBody(new JLabel(""));
    }
    
    /**
     * Add a new buttons e.g. to stop the simulation as row to the user interface.
     */
    private void addOptionButtonRow() {
        // Story details
        if (this.feeder != null) {
            JButton story = new JButton("Show story details");
            story.setName(getConfigValue("BUTTON_LABEL_PREFIX") + "story");
            story.addActionListener(this);
            addBody(story);
        } else {
            addBody(new JLabel(""));
        }
        
        // Stop simulation
        if (this.feeder != null && this.feeder.isActive()) {
            JButton stop = new JButton("Stop simulation");
            stop.setName(getConfigValue("BUTTON_LABEL_PREFIX") + "stop");
            stop.addActionListener(this);
            addBody(stop);
        } else {
            addBody(new JLabel(""));
        }
    }
    
    /**
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     * 
     * @param e                 An {@link ActionEvent}.
     */
    public final void actionPerformed(final ActionEvent e) {
        // Menu Items
        if (JMenuItem.class.isAssignableFrom(e.getSource().getClass())) {
            String menuItemType = ((JMenuItem) e.getSource()).getName();
            if ("about".equals(menuItemType)) {
                infoDialog(
                        "Version: " + MainConfig.get("REACHABILITY_UI_VERSION")
                        + "\nAuthor: " + ProjectInfo.REACHABILITY_AGENT_AUTHOR
                );
            }
        }
        
        // Buttons
        if (JButton.class.isAssignableFrom(e.getSource().getClass())) {
            JButton button = (JButton) e.getSource();
            
            // Story?
            if ((getConfigValue("BUTTON_LABEL_PREFIX") + "story").equals(button.getName())) {
                if (this.feeder != null) {
                    new ReachabilitySimulationStoryUI(this, feeder);
                }
            }
            
            // Stop?
            if ((getConfigValue("BUTTON_LABEL_PREFIX") + "stop").equals(button.getName())) {
                if (this.feeder != null && this.feeder.isActive()) {
                    this.feeder.shutDown();
                }
            }
        }
        
        // Repaint timer
        if (e.getSource().equals(this.getTimer())) {
            this.refresh(true);
        }
    }
    
    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        // Perform disconnects and notifications here...
    }
    
    /**
     * Builds the user interface.
     * 
     * @param refresh           Indicates if this method was called with the intention to re-render the user interface.
     */
    public final void refresh(final boolean refresh) {
        int rows = getFrameConfigValueAsInteger("FRAME_ROWS");
        String state = "No active simulation", path = "", file = "", speed = "", silent = "";
        if (this.feeder != null) {
            if (feeder.isActive()) {
                state = "Running";
            } else {
                state = "Finalized";
            }
            path   = feeder.getPath();
            file   = feeder.getFile();
            speed  = feeder.getSpeed() + " [smin/min]";
            silent = String.valueOf(feeder.isSilent());
            
        }

        int height = getConfigValueAsInteger("DEFAULT_WINDOW_MENU_HEIGHT")
                     + (getConfigValueAsInteger("DEFAULT_ROW_HEIGHT") * rows);
        this.setWindowHeight(height);
        fixBounds(refresh);
        
        // Define layout
        getBody().setLayout(getGrid(rows));
        getContentPane().add(getBody());
        getBody().removeAll();
        
        // Set property rows
        addTopicRow("Properties:");
        addInformationRow("State:",  state);
        addInformationRow("Path:",   path);
        addInformationRow("File:",   file);
        addInformationRow("Speed:",  speed);
        addInformationRow("Silent:", silent);
        addTopicRow("");
        
        // Set option rows
        addTopicRow("Options:");
        addOptionButtonRow();
            
        // Reload content
        fixBounds(refresh);
        getBody().revalidate();
    }
}
