package de.hawhamburg.reachability.tools.feeder;

import java.util.HashMap;
import java.util.Map;

import org.joda.time.Hours;

import de.hawhamburg.reachability.constants.NumericalConstants;
import de.hawhamburg.reachability.tools.feeder.json.JsonSensorDataFeederItem;
import de.hawhamburg.reachability.util.ExternFileUtil;

/**
 * The representation of a simulated sensor event.<br><br>
 * 
 * Objects of this class are representing simulated sensor events, which shall be send on the ActiveMQ. Each sensor 
 * event consists of a sensor type name, a time to which this event takes place and a measured value. Furthermore this
 * class provides static methods to create objects of this class from {@link String} or file and to convert it into a
 * JSON-message object. Objects of this class will be sorted by their event time.
 *
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR}
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class SensorDataFeederItem implements Comparable<SensorDataFeederItem>,
                                             Cloneable {

    /**
     * The default separator char of data within an '*.sgs' coded file.
     */
    public static final String DATA_SEPARATOR_CHAR      = ",";

    /**
     * The default separator char of lines within an '*.sgs' coded file.
     */
    public static final String LINE_SEPARATOR_CHAR      = "\\|";

    /**
     * The simulated sensor type name.
     */
    private String type                                 = null;

    /**
     * The simulated sensors measured value.
     */
    private String value                                = null;

    /**
     * The time in minutes relative to the simulation start where this event happens.
     */
    private int time                                    = 0;

    /**
     * Initializes a new {@link SensorDataFeederItem} object.
     *
     * @param typeIn        The simulated sensors type name as {@link String}.
     * @param valueIn       The simulated sensors measured value as {@link String}.
     * @param timeIn        The time when the event happens as <code>int</code>.
     */
    public SensorDataFeederItem(final String typeIn,
                                final String valueIn,
                                final int timeIn) {
        this.type  = typeIn;
        this.value = valueIn;
        this.time  = timeIn;
    }

    /**
     * Create a new {@link SensorDataFeederItem} from a given string by parsing it using {@link #DATA_SEPARATOR_CHAR} 
     * and {@link NumericalConstants#DECIMAL_BASE}.
     *
     * @param storyLine     The {@link String} line which contains the data for the simulated sensor event in "*.sgs" 
     *                      format.                      
     * @return              A new {@link SensorDataFeederItem} based on the supplied string.
     */
    public static SensorDataFeederItem fromString(final String storyLine) {
        String[] parts = storyLine.split(SensorDataFeederItem.DATA_SEPARATOR_CHAR);
        return new SensorDataFeederItem(parts[0],
                                        parts[1],
                                        Integer.parseInt(parts[2], NumericalConstants.DECIMAL_BASE));
    }
    
    /**
     * Creates all {@link SensorDataFeederItem} defined within a story file by parsing its content using 
     * {@link #LINE_SEPARATOR_CHAR} and {@link #fromString(String)} for each line.
     * 
     * @param filePath      The path to the file where the story is defined as {@link String}.
     * @return              A {@link Map} of simulation time in minutes ({@link Integer}) to 
     *                      {@link SensorDataFeederItem}.
     */
    public static Map<Integer, SensorDataFeederItem> fromFile(final String filePath) {
        Map<Integer, SensorDataFeederItem> result = new HashMap<Integer, SensorDataFeederItem>();
        
        if (ExternFileUtil.exists(filePath)) {
            for (String line : ExternFileUtil.getContentLines(filePath)) {
                for (String hiddenLine : line.split(LINE_SEPARATOR_CHAR)) {
                    if (hiddenLine.trim() != "") {
                        SensorDataFeederItem data = SensorDataFeederItem.fromString(hiddenLine);
                        result.put(data.getTime(), data);
                    }
                }
            }
        }
        
        return result;
    }
    
    /**
     * Returns the events sensor type name as {@link String}.
     * 
     * @return              The events sensor type name as {@link String}.
     */
    public final String getType() {
        return type;
    }
    
    /**
     * Returns the events simulated value as {@link String}.
     * 
     * @return              The events simulated value as {@link String}.
     */
    public final String getValue() {
        return value;
    }
    
    /**
     * Returns the event time in simulation minutes.
     * 
     * @return              The event time in simulation minutes.
     */
    public final int getTime() {
        return time;
    }
    
    /**
     * Converts the object to a JSON message representation to use in common with ActiveMQ publishing.
     * 
     * @return              A {@link JsonSensorDataFeederItem} representing the sensor event for the ActiveMQ.
     */
    public final JsonSensorDataFeederItem toJsonObject() {
        JsonSensorDataFeederItem json = new JsonSensorDataFeederItem();
        json.SensorType = type;
        json.SensorValue = value;
        return json;
    }
    
    /**
     * Returns the simulation time to which this event happens as formatted {@link String} "hh:mm".
     * 
     * @return              The simulation time to which this event happens as "hh:mm" {@link String}.
     */
    private String getRealTime() {
        return SensorDataFeederItem.getStringTime(time);
    }
    
    /**
     * @see java.lang.Object#toString()
     * 
     * @return              The {@link String} representation of this object.
     */
    public final String toString() {
        return "[" + getRealTime() + "] " + type + " (" + value + ")";
    }
    
    /**
     * Converts an amount of minutes to a simulation time {@link String} formatted as "hh:mm".
     * 
     * @param timeInMinutes The daytime in minutes as <code>int</code>.
     * @return              The daytime as {@link String} formatted as "hh:mm".
     */
    public static final String getStringTime(final int timeInMinutes) {
        int minsPerHour = Hours.ONE.toStandardMinutes().getMinutes();
        
        int minutes = Math.round(timeInMinutes);
        int mins    = minutes % minsPerHour;
        int hrs     = Math.round((minutes - mins) / minsPerHour);
        
        String result = "";
        if (hrs < NumericalConstants.DECIMAL_BASE) {
            result += "0";
        }
        result += hrs + ":";
        if (mins < NumericalConstants.DECIMAL_BASE) {
            result += "0";
        }
        result += mins;
            
        return result;
    }

    /**
     * The method compares the current object with another {@link SensorDataFeederItem} object. Both objects will be
     * compared by their event time using {@link Integer#compareTo(Integer)}.
     * 
     * @see Integer#compareTo(Integer)
     * 
     * @param other         The other {@link SensorDataFeederItem} object with which this object shall be compared. 
     * @return              <code>0</code> means that these events are starting at the same time. A value less than 
     *                      <code>0</code> indicates that this event starts before the other. A value greater than 
     *                      <code>0</code> that this event starts after the other object.
     */
    public final int compareTo(final SensorDataFeederItem other) {
        if (other == null) {
            return -1;
        }
        return Integer.valueOf(this.time).compareTo(Integer.valueOf(other.time));
    }
    
    /**
     * Clones this object and returns a new object value but not object identically to this object.
     * 
     * @return              A value but not object identical {@link SensorDataFeederItem}. 
     */
    public final SensorDataFeederItem clone() {
        return new SensorDataFeederItem(this.type, this.value, this.time);
    }
}
