package de.hawhamburg.reachability.ui.frame;

import java.awt.Cursor;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import de.hawhamburg.lifecycle.data.ConfigOrder;
import de.hawhamburg.lifecycle.data.JsonConfigOrder;
import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.conf.enumeration.ConfigType;
import de.hawhamburg.reachability.enumeration.ReachabilityGroup;
import de.hawhamburg.reachability.enumeration.ReachabilityStatus;
import de.hawhamburg.reachability.json.JsonReachabilityInstruction;
import de.hawhamburg.reachability.tools.feeder.SensorDataFeeder;
import de.hawhamburg.reachability.tools.feeder.SensorDataFeederNotifiable;
import de.hawhamburg.reachability.ui.UI;
import de.hawhamburg.reachability.ui.frame.config.ReachabilityConfigUI;
import de.hawhamburg.reachability.ui.frame.resource.ReachabilityUIUpdater;
import de.hawhamburg.reachability.ui.frame.simulation.ReachabilitySimulationDetailsUI;
import de.hawhamburg.reachability.ui.frame.simulation.ReachabilityStartSimulationUI;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.OutputUtil;

/**
 * User interface to display the actual estimated reachability states to the inhabitant.<br><br> 
 * 
 * This user interface provides full control over the reachability agent. It is able to let the agent start, stop or 
 * sleep. Furthermore the user has the ability to inform the reasoning engine, if an estimation of the engine wasn't 
 * correct in a certain case. The engine will take this in account. In addition this user interface provides the user 
 * with configuration and simulation abilities.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class ReachabilityUI extends    UI
                            implements SensorDataFeederNotifiable {

    /**
     * Some unique serial version id.
     */
    private static final long serialVersionUID                                  = -5545005309077157181L;
    
    /**
     * The {@link LPPublisher} to send the user feedback via the AMQ topic <i>REACHABILITY_AGENT_TOPIC</i>.
     */
    private LPPublisher sendUserFeedback                                        = null;
    
    /**
     * The {@link LPPublisher} to send control messages like start, stop, sleep to the reachability agent via AMQ using 
     * the topic <i>LP_CONTROL_TOPIC</i>. 
     */
    private LPPublisher controlTopic                                            = null;
    
    /**
     * The {@link ReachabilityUIUpdater} which keeps the user interface up to date.
     */
    private ReachabilityUIUpdater updater                                       = null;
    
    /**
     * The actual active simulation feeder as {@link SensorDataFeeder}.
     */
    private SensorDataFeeder simulationFeeder                                   = null;
    
    /**
     * The menu item which opens the start simulation dialog.
     */
    private JMenuItem startSimulationMenuItem                                   = null;
    
    /**
     * The menu item which opens the simulation control dialog.
     */
    private JMenuItem showSimulationMenuItem                                    = null;
    
    /**
     * The current known reachability state.
     */
    private Map<ReachabilityGroup, ReachabilityStatus> status                   =
            new HashMap<ReachabilityGroup, ReachabilityStatus>();
    
    /**
     * The different selectors ({@link JList}s) for each {@link ReachabilityGroup} as {@link Map} for refreshing.
     */
    private Map<ReachabilityGroup, JList<ReachabilityStatus>> stateSelectors    =
            new HashMap<ReachabilityGroup, JList<ReachabilityStatus>>();
    
    /**
     * Creates and builds a new reachability user interface.
     */
    public ReachabilityUI() {
        super("Reachability Status");
        
        String feedbackTopicName    = MainConfig.get("REACHABILITY_AGENT_TOPIC");
        String controlTopicName     = MainConfig.get("LP_CONTROL_TOPIC");
        
        sendUserFeedback            = ConnectionUtil.getPublisher(feedbackTopicName, this);
        controlTopic                = ConnectionUtil.getPublisher(controlTopicName, this);
        
        updater = new ReachabilityUIUpdater(this);
        updater.start();
        
        this.build();
    }
    
    /**
     * This method adds a menu to the user interface.
     */
    protected final void addMenu() {
        JMenu control = new JMenu("Control");
        control.setMnemonic('C');
        
        JMenuItem start = new JMenuItem("Start");
        start.setMnemonic('S');
        start.setName("start");
        control.add(start);
        
        JMenuItem sleep = new JMenuItem("Sleep");
        sleep.setMnemonic('e');
        sleep.setName("sleep");
        control.add(sleep);
        
        JMenuItem stop = new JMenuItem("Stop");
        stop.setMnemonic('t');
        stop.setName("stop");
        control.add(stop);
        
        JMenu config = new JMenu("Configuration");
        config.setMnemonic('f');
        
        JMenuItem system = new JMenuItem("System");
        system.setMnemonic('y');
        system.setName("system");
        config.add(system);
        
        JMenuItem reasoner = new JMenuItem("Reasoner");
        reasoner.setMnemonic('R');
        reasoner.setName("reasoner");
        config.add(reasoner);
        
        JMenu simulation = new JMenu("Simulation");
        simulation.setMnemonic('i');
        
        startSimulationMenuItem = new JMenuItem("Start simulation");
        startSimulationMenuItem.setMnemonic('s');
        startSimulationMenuItem.setName("startSimulation");
        simulation.add(startSimulationMenuItem);
        
        showSimulationMenuItem = new JMenuItem("Show simulation details");
        showSimulationMenuItem.setMnemonic('d');
        showSimulationMenuItem.setName("showSimulationDetails");
        showSimulationMenuItem.setEnabled(false);
        simulation.add(showSimulationMenuItem);
        
        JMenu help = new JMenu("Help");
        help.setMnemonic('H');
        
        JMenuItem about = new JMenuItem("About Reachability Agent");
        about.setMnemonic('A');
        about.setName("about");
        help.add(about);
        
        JMenuBar bar = new JMenuBar();
        bar.add(control);
        bar.add(config);
        bar.add(simulation);
        bar.add(help);
        setJMenuBar(bar);
        
        // Actions
        start.addActionListener(this);
        sleep.addActionListener(this);
        stop.addActionListener(this);
        system.addActionListener(this);
        reasoner.addActionListener(this);
        startSimulationMenuItem.addActionListener(this);
        showSimulationMenuItem.addActionListener(this);
        about.addActionListener(this);
    }
    
    /**
     * Returns a user interface grid.
     * 
     * @return              Returns a user interface grid.
     */
    private GridLayout getGrid() {
        return getGrid(2, ReachabilityGroup.values().length);
    }
    
    /**
     * Adds a title row to the user interface containing the names of the reachability groups
     * ({@link ReachabilityGroup}).
     */
    private void addTitleRow() {
        for (ReachabilityGroup grp : ReachabilityGroup.values()) {
            addBody(new JLabel(grp.toString()));
        }
    }
    
    /**
     * Updates the status rows defined by {@link #stateSelectors} with new values.
     */
    private void updateStatusRow() {
        for (ReachabilityGroup grp : ReachabilityGroup.values()) {
            // Get selected status
            ReachabilityStatus selected = ReachabilityStatus.NOT_CALCULATED;
            if (status.containsKey(grp)) {
                selected = status.get(grp);
            }
            
            // Set selected status
            if (stateSelectors.containsKey(grp)) {
                stateSelectors.get(grp).setSelectedValue(selected, true);
            } else {
                JList<ReachabilityStatus> list = getStatusList();
                list.setCursor(new Cursor(Cursor.HAND_CURSOR));
                list.setName(getConfigValue("FIELD_LABEL_PREFIX_JLIST") + grp.toString());
                list.setSelectedValue(selected, true);
                list.addMouseListener(this);
                addBody(list);
                stateSelectors.put(grp, list);
            }
        }
    }
    
    /**
     * Returns a {@link JList} of all {@link ReachabilityStatus} entries.
     * 
     * @return              A {@link JList} containing all {@link ReachabilityStatus} entries as items.
     */
    private JList<ReachabilityStatus> getStatusList() {
        ReachabilityStatus[] items = ReachabilityStatus.values();
        return new JList<ReachabilityStatus>(items);
    }
    
    /**
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     * 
     * @param e             An {@link ActionEvent}.
     */
    public final void actionPerformed(final ActionEvent e) {
        String menuItemType = ((JMenuItem) e.getSource()).getName();
        if ("about".equals(menuItemType)) {
            infoDialog("Version: " + MainConfig.get("REACHABILITY_UI_VERSION")
                       + "\nAuthor: " + ProjectInfo.REACHABILITY_AGENT_AUTHOR);
        }
        if ("start".equals(menuItemType) || "sleep".equals(menuItemType) || "stop".equals(menuItemType)) {
            sendControl(ConfigOrder.valueOf(menuItemType));
        }
        if ("system".equals(menuItemType)) {
            configDialog(ConfigType.SYSTEM);
        }
        if ("reasoner".equals(menuItemType)) {
            configDialog(ConfigType.REASONER);
        }
        if ("startSimulation".equals(menuItemType)) {
            startSimulationDialog();
        }
        if ("showSimulationDetails".equals(menuItemType)) {
            showSimulationDetailsDialog();
        }
    }

    /**
     * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
     * 
     * @param e             A {@link MouseEvent}.
     */
    public final void mousePressed(final MouseEvent e) {
        JComponent src = (JComponent) e.getSource();
        String     key = getConfigValue("FIELD_LABEL_PREFIX_JLIST");
        
        if (key.equals(src.getName().substring(0, key.length()))) {
            
            @SuppressWarnings("unchecked")
            JList<ReachabilityStatus> list      = (JList<ReachabilityStatus>) e.getSource();
            ReachabilityGroup group             = ReachabilityGroup.valueOf(list.getName().substring(key.length()));
            ReachabilityStatus theStatus        = (ReachabilityStatus) list.getSelectedValue();
            JsonReachabilityInstruction inst    = getInstructionMsg(group, theStatus);
            
            ConnectionUtil.setMessageOnTopic(sendUserFeedback, inst);
        }
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        // Shut down other threads
        if (updater != null) {
            updater.shutDown();
        }
        
        // Shut down own AMQ publisher/subscriber
        if (sendUserFeedback != null) {
            sendUserFeedback.disconnect();
        }
        if (controlTopic != null) {
            controlTopic.disconnect();
        }
    }
    
    /**
     * Open a new dialog to start a simulation.
     */
    private void startSimulationDialog() {
        new ReachabilityStartSimulationUI(this);
    }
    
    /**
     * Opens a new dialog to show the current simulation state.
     */
    private void showSimulationDetailsDialog() {
        new ReachabilitySimulationDetailsUI(this, simulationFeeder);
    }
    
    /**
     * Show the configuration dialog.
     * 
     * @param type          The type of configuration.
     */
    private void configDialog(final ConfigType type) {
        new ReachabilityConfigUI(this, type);
    }
    
    /**
     * Initial build of this user interface. The rendering won't be influenced by the <i>refresh</i> parameter.
     * 
     * @param refresh       Indicates if this method was called with the intention to re-render the interface.
     */
    public final void refresh(final boolean refresh) {
        refresh(null);
    }
    
    /**
     * Builds/refreshes the cached {@link ReachabilityStatus} and the user interface.
     * 
     * @param newStatus     A hash {@link Map} containing the current {@link ReachabilityStatus}.
     */
    public final void refresh(final Map<ReachabilityGroup, ReachabilityStatus> newStatus) {
        if (newStatus != null) {
            this.status = newStatus;
        } else {
            getBody().setLayout(getGrid());
            getContentPane().add(getBody());
            getBody().removeAll();
            addTitleRow();
        }
        updateStatusRow();
        getBody().revalidate();
    }
    
    /**
     * Calls {@link #getInstructionMsg(ReachabilityGroup, ReachabilityStatus, String)} with a blank {@link String} as 
     * remark and returns its result.
     * 
     * @param group         A {@link ReachabilityGroup}.
     * @param resStatus     A {@link ReachabilityStatus}
     * @return              A {@link JsonReachabilityInstruction} message.
     */
    public final JsonReachabilityInstruction getInstructionMsg(final ReachabilityGroup group,
                                                               final ReachabilityStatus resStatus) {
        return getInstructionMsg(group, resStatus, "");
    }
    
    /**
     * Generates a {@link JsonReachabilityInstruction} messages and returns it.
     * 
     * @param group         A {@link ReachabilityGroup}
     * @param resStatus     A {@link ReachabilityStatus}
     * @param remark        A remark as {@link String}.
     * @return              A {@link JsonReachabilityInstruction} message.
     */
    public final JsonReachabilityInstruction getInstructionMsg(final ReachabilityGroup group,
                                                               final ReachabilityStatus resStatus,
                                                               final String remark) {
        JsonReachabilityInstruction inst = new JsonReachabilityInstruction();
        inst.Id         = getId();
        inst.Version    = MainConfig.get("REACHABILITY_UI_VERSION");
        inst.Group      = group;
        inst.Status     = resStatus;
        inst.Remark     = remark;
        return inst;
    }
    
    /**
     * Sends a {@link ConfigOrder} in a {@link JsonConfigOrder} messages via AMQ to the reachability agent.
     * 
     * @param x             The {@link ConfigOrder} which shall be send.
     */
    public final void sendControl(final ConfigOrder x) {
        JsonConfigOrder order = new JsonConfigOrder();
        order.Version   = MainConfig.get("REACHABILITY_UI_VERSION");
        order.Id        = ProjectInfo.REACHABILITY_AGENT_NAME;
        order.order     = x;
        ConnectionUtil.setMessageOnTopic(controlTopic, order);
    }

    /**
     * Registers a new simulation feeder. Disables the start simulation and enables the simulation controlling menu 
     * option.
     * 
     * @param feeder        The new simulation feeder as {@link SensorDataFeeder} which shall be registered.
     */
    public final void registerSimulationFeeder(final SensorDataFeeder feeder) {
        this.simulationFeeder = feeder;
        this.startSimulationMenuItem.setEnabled(false);
        this.showSimulationMenuItem.setEnabled(true);
    }

    /**
     * Unregisters a simulation feeder. Disables the simulation controlling and enables the start start simulation menu
     * option.
     * 
     * @param feeder        The simulation feeder as {@link SensorDataFeeder} which shall be unregistered.
     */
    public final void unregisterSimulationFeeder(final SensorDataFeeder feeder) {
        if (this.simulationFeeder.equals(feeder)) {
            this.simulationFeeder = null;
            this.startSimulationMenuItem.setEnabled(true);
            this.showSimulationMenuItem.setEnabled(false);
        }
    }

    /**
     * Starts a new {@link ReachabilityUI} in the systems look and feel.
     * 
     * @param args              The system supplied start arguments as array of {@link String}.
     */
    public static void main(final String[] args) {
        // Remove all existing loggers
        org.apache.log4j.Logger.getRootLogger().removeAllAppenders();
        
        // Try to adapt the systems look and feel
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (ClassNotFoundException e) {
            OutputUtil.error("[ClassNotFoundException] " + e.getMessage());
        } catch (InstantiationException e) {
            OutputUtil.error("[InstantiationException] " + e.getMessage());
        } catch (IllegalAccessException e) {
            OutputUtil.error("[IllegalAccessException] " + e.getMessage());
        } catch (UnsupportedLookAndFeelException e) {
            OutputUtil.error("[UnsupportedLookAndFeelException] " + e.getMessage());
        }
        
        // Start a new color user interface
        new ReachabilityUI();
    }
}
