package de.hawhamburg.reachability.ui.frame.config;

import java.awt.Cursor;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;

import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.conf.enumeration.ConfigType;
import de.hawhamburg.reachability.properties.DisplayType;
import de.hawhamburg.reachability.properties.PropertyDescription;
import de.hawhamburg.reachability.properties.json.JsonRAPropertiesChange;
import de.hawhamburg.reachability.properties.json.JsonRAPropertiesRequest;
import de.hawhamburg.reachability.ui.UI;
import de.hawhamburg.reachability.ui.frame.config.resource.ReachabilityConfigUIResultListener;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.OutputUtil;

/**
 * User interface to adjust the reachability agents parameters.<br><br> 
 * 
 * This user interface is a piggyback window to display and change the current configuration of the reachability agent. 
 * It is able to load the current properties via ActiveMQ from the reachability agent and present them to the user. The 
 * user then may change the configuration to influence the calculation of the reachability and to adjust the 
 * reachability processing to be more comfortable for him.
 *
 * This user interface represents all possible configuration user interfaces independent from the chosen properties 
 * type. Possible types are defined by the {@link ConfigType} enumeration.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class ReachabilityConfigUI extends UI {

    /**
     * Some unique serial version id.
     */
    private static final long serialVersionUID              = -5545005309077157181L;
    
    /**
     * The content pane of this user interface as {@link JPanel}.
     */
    private JPanel contentPane                              = null;
    
    /**
     * The type of configuration pane.
     */
    private ConfigType internalType                         = null;
    
    /**
     * The current service state reference.
     */
    private String serviceStateReference                    = null;
        
    /**
     * The last known configuration.
     */
    private Map<String, String> config                      = null;
    
    /**
     * The last known property description.
     */
    private Map<String, PropertyDescription> propertyDesc   = null;
    
    /**
     * Current active updater.
     */
    private Set<ReachabilityConfigUIResultListener> updaters;
    
    /**
     * The {@link LPPublisher} to send the users configuration via the ActiveMQ topic 
     * <i>REACHABILITY_CONFIG_CHANGE_TOPIC</i>.
     */
    private LPPublisher changeConfig                        = null;
    
    /**
     * The {@link LPPublisher} to send the get configuration request <i>REACHABILITY_CONFIG_REQUEST_TOPIC</i>.
     */
    private LPPublisher requestConfig                       = null;
        
    /**
     * Initializes, starts and builds a new {@link ReachabilityConfigUI} which provides an overview over a certain area 
     * of configuration parameters related to the reachability agent. The user is given also the ability to change these
     * parameters and therefore adapt the system to his or her personal preferences. Furthermore it opens the ActiveMQ 
     * connections for communicating with the reachability agent by using JSON messages.
     * 
     * @param parentWindow  The window from which this interface was created as {@link JFrame}.
     * @param type          The type of configuration parameters. One out of {@link ConfigType}.
     */
    public ReachabilityConfigUI(final JFrame parentWindow,
                                final ConfigType type) {
        super("Reachability Config: " + type, parentWindow);
        this.internalType       = type;
        
        // Fetch required topic names
        String requestTopic = MainConfig.get("REACHABILITY_CONFIG_REQUEST_TOPIC");
        String changeTopic  = MainConfig.get("REACHABILITY_CONFIG_CHANGE_TOPIC");
        
        // Set up publishers
        requestConfig = ConnectionUtil.getPublisher(requestTopic, this);
        changeConfig  = ConnectionUtil.getPublisher(changeTopic, this);
        
        // Build the interface
        build();
    }
    
    /**
     * This method adds a menu to the user interface.
     */
    protected final void addMenu() {
        if (isStandalone()) {
            JMenu help = new JMenu("Help");
            help.setMnemonic('H');
            
            JMenuItem about = new JMenuItem("About Reachability Agent");
            about.setMnemonic('A');
            about.setName("about");
            help.add(about);
            
            JMenuBar bar = new JMenuBar();
            bar.add(help);
            setJMenuBar(bar);
            
            // Actions
            about.addActionListener(this);
        }
    }
    
    /**
     * Returns a user interface grid.
     * 
     * @param rowsIn        Amount of rows to be displayed
     * @return              Returns a user interface grid.
     */
    private GridLayout getGrid(final int rowsIn) {
        return getGrid(rowsIn, 2);
    }
    
    /**
     * Adds a new row to the users interface, containing the properties type, its value and the possibility to change 
     * it.
     * 
     * @param key           The properties name as {@link String}.
     * @param desc          The display parameters of the property as {@link PropertyDescription}.
     */
    private void addConfigRow(final String key,
                              final PropertyDescription desc) {
        String selected             = null;
        if (config.containsKey(key)) {
            selected = config.get(key);
        }
        
        DisplayType dspType = DisplayType.ENUM;
        if (desc != null) {
            dspType = desc.getDisplayType();
            contentPane.add(new JLabel(desc.getDisplayName()));
        } else {
            contentPane.add(new JLabel(key));
        }
        
        switch (dspType) {
            
            case TEXTFIELD:
                if (selected == null) {
                    selected = "";
                }
                JTextArea area = new JTextArea(selected,
                                               getConfigValueAsInteger("DEFAULT_TEXTAREA_ROWS"),
                                               getConfigValueAsInteger("DEFAULT_TEXTAREA_COLUMNS"));
                area.setEditable(true);
                area.setName(getConfigValue("FIELD_LABEL_PREFIX_JTEXTAREA") + key);
                contentPane.add(area);
                break;
                
            case INT:
                if (selected == null) {
                    selected = "0";
                }
                JTextField field = new JTextField(selected);
                field.setEditable(true);
                field.setName(getConfigValue("FIELD_LABEL_PREFIX_JTEXTFIELD_INT") + key);
                field.addKeyListener(this);
                contentPane.add(field);
                break;
                
            case ENUM:
            case AGENT_DEFINED:
            default:
                int size = 0;
                if (desc != null) {
                    size = desc.getPossibleValue().size();
                }
                String convertedSelected = selected;
                if (desc != null && desc.getPossibleValue().containsKey(selected)) {
                    convertedSelected = desc.getPossibleValue().get(selected);
                } else {
                    size++;
                }
                
                String[] convertedOptions = new String[size];
                int c = 0;
                if (desc != null) {
                    for (String converted : desc.getPossibleValue().values()) {
                        convertedOptions[c++] = converted;
                    }
                }
                if (size != c) {
                    convertedOptions[c] = convertedSelected;
                }

                JComboBox<String> list = new JComboBox<String>(convertedOptions);
                list.setCursor(new Cursor(Cursor.HAND_CURSOR));
                list.setName(getConfigValue("FIELD_LABEL_PREFIX_JCOMBOBOX") + key);
                if (convertedOptions != null) {
                    list.setSelectedItem(convertedSelected);
                }
                list.addActionListener(this);
                contentPane.add(list);
        }
    }
    
    /**
     * Adds a new topic row to the user interface.
     * 
     * @param topic         The topics name as {@link String}.
     */
    private void addTopicRow(final String topic) {
        JLabel top      = new JLabel(topic);
        Font parentFont = getBody().getFont();
        top.setFont(new Font(parentFont.getFamily(), Font.BOLD, parentFont.getSize()));
        contentPane.add(top);
        contentPane.add(new JLabel(""));
    }
    
    /**
     * Add a new button to submit the data as row to the user interface.
     */
    private void addSubmitRow() {
        // Refresh button
        contentPane.add(new JLabel(""));
        
        // Submit button
        JButton submit = new JButton("Save");
        submit.setName("JButton_submit");
        submit.addActionListener(this);
        contentPane.add(submit);
    }
    
    /**
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     * 
     * @param e             An {@link ActionEvent}.
     */
    public final void actionPerformed(final ActionEvent e) {
        // Menu Items
        if (JMenuItem.class.isAssignableFrom(e.getSource().getClass())) {
            String menuItemType = ((JMenuItem) e.getSource()).getName();
            if ("about".equals(menuItemType)) {
                infoDialog(
                        "Version: " + MainConfig.get("REACHABILITY_UI_VERSION")
                        + "\nAuthor: " + ProjectInfo.REACHABILITY_AGENT_AUTHOR
                );
            }
        }
        
        // Selectbox
        if (JComboBox.class.isAssignableFrom(e.getSource().getClass())) {
            Map<String, String> props = new HashMap<String, String>();
            
            @SuppressWarnings("unchecked")
            JComboBox<String> list = (JComboBox<String>) e.getSource();
            props.putAll(fetchFromComboxBox(list));
            sendConfigChange(props);
        }
        
        // Buttons
        if (JButton.class.isAssignableFrom(e.getSource().getClass())) {
            JButton button = (JButton) e.getSource();
            if ("JButton_submit".equals(button.getName())) {
                Map<String, String> props = new HashMap<String, String>();
                for (int i = 0; i < contentPane.getComponentCount(); i++) {
                    // Selectbox found
                    if (JComboBox.class.isAssignableFrom(contentPane.getComponent(i).getClass())) {
                        @SuppressWarnings("unchecked")
                        JComboBox<String> list = (JComboBox<String>) contentPane.getComponent(i);
                        props.putAll(fetchFromComboxBox(list));
                    }
                    
                    // TextArea found
                    if (JTextArea.class.isAssignableFrom(contentPane.getComponent(i).getClass())) {
                        JTextArea textarea = (JTextArea) contentPane.getComponent(i);
                        props.putAll(fetchFromTextArea(textarea));
                    }
                    
                    // TextField found
                    if (JTextField.class.isAssignableFrom(contentPane.getComponent(i).getClass())) {
                        JTextField textfield = (JTextField) contentPane.getComponent(i);
                        props.putAll(fetchFromTextField(textfield));
                    }
                }
                sendConfigChange(props);
            }
        }
    }

    /**
     * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
     * 
     * @param arg0          A {@link KeyEvent}.
     */
    public final void keyReleased(final KeyEvent arg0) {
        // JTextFields
        if (JTextField.class.isAssignableFrom(arg0.getSource().getClass())) {
            JTextField field = (JTextField) arg0.getSource();
            
            // Integer only
            String fieldLabel = getConfigValue("FIELD_LABEL_PREFIX_JTEXTFIELD_INT");
            if (field.getName().startsWith(fieldLabel)) {
                // Check content
                field.setText(field.getText().replaceAll("[^0-9-]", ""));
                
                // Check range
                String   key    = field.getName().substring(fieldLabel.length());
                try {
                    Integer value = Integer.valueOf(field.getText());
                    if (this.propertyDesc.containsKey(key)) {
                        Integer min     = Integer.valueOf(this.propertyDesc.get(key).getPossibleValue().get("min"));
                        Integer max     = Integer.valueOf(this.propertyDesc.get(key).getPossibleValue().get("max"));
                        
                        if (value < min || value > max) {
                            infoDialog("Invalid value supplied. It needs to be a value between " + min + " and " + max);
                            field.setText("0");
                        }
                    }
                    
                } catch (NumberFormatException e) {
                    OutputUtil.error(new Exception("").getStackTrace()[0].getMethodName()
                                     + "[NumberFormatException] " + e.getMessage());
                }
            }
        }
    }
    
    /**
     * Fetches the properties value from a given {@link JComboBox} and returns it as {@link Map} of {@link String} to 
     * {@link String}.
     * 
     * @param list          The {@link JComboBox} object including options as list of {@link String}.
     * @return              The properties value fetched from the given {@link JComboBox} as {@link Map} of 
     *                      {@link String} to {@link String}.
     */
    private Map<String, String> fetchFromComboxBox(final JComboBox<String> list) {
        String key      = getConfigValue("FIELD_LABEL_PREFIX_JCOMBOBOX");
        String property = list.getName().substring(key.length());
        String value    = (String) list.getSelectedItem();
        if (propertyDesc.containsKey(property) && propertyDesc.get(property).getPossibleValue().containsValue(value)) {
            for (Map.Entry<String, String> pair : propertyDesc.get(property).getPossibleValue().entrySet()) {
                if (pair.getValue().equals(value)) {
                    value = pair.getKey();
                    break;
                }
            }
        }
        
        Map<String, String> result = new HashMap<String, String>();
        result.put(property, value);
        return result;
    }
    
    /**
     * Fetches the properties value from a given {@link JTextArea} and returns it as {@link Map} of {@link String} to 
     * {@link String}.
     *  
     * @param textarea      The {@link JTextArea} object including the properties new value as 
     *                      {@link String}.
     * @return              The properties value fetched from the given {@link JTextArea} as {@link Map} of 
     *                      {@link String} to {@link String}.
     */
    private Map<String, String> fetchFromTextArea(final JTextArea textarea) {
        String key      = getConfigValue("FIELD_LABEL_PREFIX_JTEXTAREA");
        String property = textarea.getName().substring(key.length());
        String value    = (String) textarea.getText();

        Map<String, String> result = new HashMap<String, String>();
        result.put(property, value);
        return result;
    }
    
    /**
     * Fetches the properties value from a given {@link JTextField} and returns it as {@link Map} of {@link String} to 
     * {@link String}.
     *  
     * @param textfield     The {@link JTextField} object including the properties new value as 
     *                      {@link String}.
     * @return              The properties value fetched from the given {@link JTextField} as {@link Map} of 
     *                      {@link String} to {@link String}.
     */
    private Map<String, String> fetchFromTextField(final JTextField textfield) {
        String keyText = getConfigValue("FIELD_LABEL_PREFIX_JTEXTFIELD");
        String keyInt  = getConfigValue("FIELD_LABEL_PREFIX_JTEXTFIELD_INT");
        String property;
        if (textfield.getName().startsWith(keyInt)) {
            property = textfield.getName().substring(keyInt.length());
        } else {
            property = textfield.getName().substring(keyText.length());
        }
        String value    = (String) textfield.getText();

        Map<String, String> result = new HashMap<String, String>();
        result.put(property, value);
        return result;
    }
    
    /**
     * Sends a new {@link JsonRAPropertiesChange} message to change one or more properties to the certain 
     * {@link ConfigType} by using {@link #getChangeMsg(Map)} and the {@link ConnectionUtil}.
     * 
     * @param changed       The changed properties as {@link Map} of {@link String} to {@link String} which contains
     *                      the property names and their new values.
     */
    private void sendConfigChange(final Map<String, String> changed) {
        JsonRAPropertiesChange chg = getChangeMsg(changed);
        ConnectionUtil.setMessageOnTopic(changeConfig, chg);
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        // Shut down all active updaters
        for (ReachabilityConfigUIResultListener updater : updaters) {
            updater.shutDown();
        }
        
        // Shut down own AMQ publisher/subscriber
        if (requestConfig != null) {
            requestConfig.disconnect();
        }
        if (changeConfig != null) {
            changeConfig.disconnect();
        }
    }
    
    /**
     * Builds the user interface. The rendering won't be influenced by the <i>refresh</i> parameter.
     * 
     * @param refresh       Indicates if this method has been called with the intention to re-build the user interface.
     */
    public final void refresh(final boolean refresh) {
        refresh(null, null, null, null);
    }
    
    /**
     * Fills the panel with (new) informations. Supervises the rendering of the panels content. The method might be only
     * called by the {@link #refresh(boolean)} method or an {@link ReachabilityConfigUIResultListener} for actual 
     * configuration informations on refresh.
     * 
     * @param props         A {@link Map} of {@link String} to {@link String} which shall be 
     *                      rendered changeable as content.
     * @param desc          A {@link Map} of {@link String} to {@link Map} of {@link String} to 
     *                      {@link PropertyDescription} which describes the grouping, output texts and order of the
     *                      properties.
     * @param ssr           The service state reference for change requests as {@link String}.
     * @param updater       The updater object to be removed from the list of known updater threads
     *                      ({@link #updaters}).
     */
    public final void refresh(final Map<String, String> props,
                              final Map<String, Map<String, PropertyDescription>> desc,
                              final String ssr,
                              final ReachabilityConfigUIResultListener updater) {
        this.serviceStateReference = ssr;
        
        int rows    = 1;
        boolean fnd = false;
        if (props != null && !props.isEmpty()) {
            updaters.remove(updater);
            
            if (desc != null) {
                for (Map<String, PropertyDescription> descr : desc.values()) {
                    if (this.propertyDesc == null) {
                        this.propertyDesc = new HashMap<String, PropertyDescription>();
                    }
                    this.propertyDesc.putAll(descr);
                }
            }
            
            fnd                 = true;
            this.config         = props;
            int addRows         = desc.size() * 2;
            rows                = config.size() + addRows + 1;
            int newHeight       = getConfigValueAsInteger("DEFAULT_WINDOW_MENU_HEIGHT")
                                + (getConfigValueAsInteger("DEFAULT_ROW_HEIGHT") * rows);
            int maxHeight       = getConfigValueAsInteger("MAX_DISPLAY_HEIGHT");
            if (newHeight > maxHeight) {
                newHeight       = maxHeight;
            }
            
            setWindowHeight(newHeight);
            fixBounds(true);
        } else if (props != null) {
            fnd = true;
        }

        // Define layout
        getBody().setLayout(getGrid(1, 1));
        contentPane = new JPanel();
        contentPane.setLayout(getGrid(rows));
        contentPane.setVisible(true);
        
        JScrollPane scroll = new JScrollPane(contentPane);
        scroll.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
        scroll.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        scroll.getVerticalScrollBar().setUnitIncrement(getConfigValueAsInteger("SCROLL_SPEED"));
        scroll.setVisible(true);
        
        getBody().removeAll();
        getBody().add(scroll);
        
        getContentPane().add(getBody());
        contentPane.removeAll();
        
        // Set configuration rows
        if (config != null) {
            for (Map.Entry<String, Map<String, PropertyDescription>> group : desc.entrySet()) {
                addTopicRow(group.getKey());
                
                for (Map.Entry<String, PropertyDescription> descEntry : group.getValue().entrySet()) {
                    if (config.containsKey(descEntry.getKey())) {
                        addConfigRow(descEntry.getKey(), descEntry.getValue());
                    }
                }
                
                addTopicRow("");
            }
            
            addSubmitRow();
            
            // Reload content
            fixBounds(true);
        } else {
            String story = "Loading...";
            if (fnd) {
                story += "(Der Server braucht ungewöhnlich lange, es wird erneut versucht)";
            }
            contentPane.add(new JLabel(story));
        }
        contentPane.revalidate();
        getBody().revalidate();
        
        if (!fnd) {
            requestConfiguration();
        }
    }
    
    /**
     * Requests the current configuration for the actual {@link ConfigType} by sending a JSON message on the AMQ request
     * topic. Additionally starts a new {@link ReachabilityConfigUIResultListener} thread to listen for the response.
     */
    private void requestConfiguration() {
        if (this.updaters == null) {
            this.updaters = new HashSet<ReachabilityConfigUIResultListener>();
        }
        
        if (this.updaters.isEmpty()) {
            JsonRAPropertiesRequest req = getRequestMsg();
            ReachabilityConfigUIResultListener updater = new ReachabilityConfigUIResultListener(this, req.Id);
            updater.start();
            try {
                Thread.sleep(getConfigValueAsInteger("REQUEST_WAITING"));
            } catch (InterruptedException e) {
                OutputUtil.error("No sleeping time: " + e.getMessage());
            }
            ConnectionUtil.setMessageOnTopic(requestConfig, req);
            this.updaters.add(updater);
        }
    }
    
    /**
     * Builds up a {@link JsonRAPropertiesChange} message to change the values of certain properties.
     * 
     * @param properties    The properties which shall be changed as {@link Map} of property to value. 
     * @return              The {@link JsonRAPropertiesChange} message.
     */
    public final JsonRAPropertiesChange getChangeMsg(final Map<String, String> properties) {
        JsonRAPropertiesChange chg = new JsonRAPropertiesChange();
        
        chg.Id                      = getId();
        chg.Version                 = MainConfig.get("REACHABILITY_UI_VERSION");
        chg.Type                    = this.internalType;
        chg.ServiceStateReference   = this.serviceStateReference;
        chg.Properties              = properties;
        
        return chg;
    }
    
    /**
     * Builds up a {@link JsonRAPropertiesRequest} message to request the current settings for the actual 
     * {@link ConfigType}.
     *  
     * @return              The {@link JsonRAPropertiesRequest} message for all properties of the current 
     *                      {@link ConfigType}.
     */
    public final JsonRAPropertiesRequest getRequestMsg() {
        JsonRAPropertiesRequest req = new JsonRAPropertiesRequest();
        req.Id          = getId();
        req.Version     = MainConfig.get("REACHABILITY_UI_VERSION");
        req.Type        = this.internalType;
        req.Property    = "*ALL";
        return req;
    }
}
