package de.hawhamburg.reachability.ui.frame.simulation;

import java.awt.Container;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.util.SortedMap;
import java.util.TreeMap;

import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.tools.feeder.SensorDataFeeder;
import de.hawhamburg.reachability.tools.feeder.SensorDataFeederItem;
import de.hawhamburg.reachability.ui.UI;

/**
 * Displays all sent events of the current simulation defined by the supplied simulation feeder to the user.<br><br>
 * 
 * This user interfaces displays all sent events of the supplied simulation feeder. It gives the user the ability to
 * inspect the simulation more detailed.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class ReachabilitySimulationStoryUI extends UI {
    
    /**
     * Some unique serial version id.
     */
    private static final long serialVersionUID = 1032956844537719373L;
    
    /**
     * The active simulation feeder as {@link SensorDataFeeder}.
     */
    private SensorDataFeeder feeder             = null;
    
    /**
     * The last scroll bar index the user has focused.
     */
    private int lastUserScrollIndex             = 0;
    
    /**
     * The last time the user has used the scroll bar function.
     */
    private long lastUserScrollTime             = 0;
    
    /**
     * The {@link JTextArea} which holds the displayed story.
     */
    private JTextArea story                     = null;
    
    /**
     * Creates and builds a new {@link ReachabilitySimulationStoryUI} which displays more detailed the sended
     * events of the running simulation.
     * 
     * @param sourceWindow  The calling panel which will be notified, when the simulation was
     *                      started as Object implementing {@link JFrame}.
     * @param runningFeeder The active simulation feeder as {@link SensorDataFeeder}.
     */
    public ReachabilitySimulationStoryUI(final JFrame sourceWindow,
                                         final SensorDataFeeder runningFeeder) {
        super("Simulation story", sourceWindow);
        this.feeder = runningFeeder;
        build();
    }
    
    /**
     * Creates and builds a new {@link ReachabilitySimulationStoryUI} by using 
     * {@link #ReachabilitySimulationStoryUI(JFrame, SensorDataFeeder)} without calling panel.
     * 
     * @param runningFeeder The active simulation feeder as {@link SensorDataFeeder}.
     */
    public ReachabilitySimulationStoryUI(final SensorDataFeeder runningFeeder) {
        this(null, runningFeeder);
    }
    
    /**
     * This method adds a menu to the user interface.
     */
    protected final void addMenu() {
        if (this.isStandalone()) {
            JMenu help = new JMenu("Help");
            help.setMnemonic('H');
            
            JMenuItem about = new JMenuItem("About Reachability Agent");
            about.setMnemonic('A');
            about.setName("about");
            help.add(about);
            
            JMenuBar bar = new JMenuBar();
            bar.add(help);
            setJMenuBar(bar);
            
            // Actions
            about.addActionListener(this);
        }
    }
    
    /**
     * Returns a user interface grid.
     * 
     * @param rowsIn        Amount of rows to be displayed
     * @return              Returns a user interface grid.
     */
    private GridLayout getGrid(final int rowsIn) {
        return getGrid(1, 1);
    }
    
    /**
     * Refreshes the content of the {@link #story} by the actual <i>storyMap</i> and handles correct scroll position of
     * this panel afterwards.
     * 
     * @param storyMap      The actual sent events as {@link SortedMap} of {@link SensorDataFeederItem} to 
     *                      {@link String}.
     */
    private void addStoryRow(final SortedMap<SensorDataFeederItem, String> storyMap) {
        // Generate the stories text
        String storyText = "Loading...";
        boolean done = false;
        for (SortedMap.Entry<SensorDataFeederItem, String> entry : storyMap.entrySet()) {
            if (!done) {
                storyText = "";
            }
            storyText += "Topic: " + entry.getValue() + " | Event: " + entry.getKey() + "\n";
            done = true;
        }
        
        // Does the textarea container exist?
        if (story == null) {
            // Build new textarea
            story = new JTextArea(storyText);
            story.setEditable(false);
            
            // Initialize scroll panel container for textarea
            JScrollPane logContainer = new JScrollPane(story);
            logContainer.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
            logContainer.getVerticalScrollBar().addAdjustmentListener(
                new AdjustmentListener() {
                    public void adjustmentValueChanged(final AdjustmentEvent e) {
                        boolean userAction = e.getValueIsAdjusting();
                        int value = e.getAdjustable().getMaximum();
                        if (JScrollBar.class.isAssignableFrom(e.getSource().getClass())) {
                            
                            // Get ReachabilitySimulationStoryUI
                            JScrollBar x = (JScrollBar) e.getSource();
                            Container frame = x.getRootPane().getParent();
                            if (ReachabilitySimulationStoryUI.class.
                                    isAssignableFrom(frame.getClass())) {
                                ReachabilitySimulationStoryUI ui =
                                        (ReachabilitySimulationStoryUI) frame;
                                if (userAction) {
                                    ui.updateUserScrollAction(e.getValue());
                                } else {
                                    int mind  = getConfigValueAsInteger("MIND_USER_SCROLL_ACTION");
                                    long diff = System.currentTimeMillis()
                                                - ui.getLastUserScrollActionTime();
                                    if (diff < mind) {
                                        value = ui.getLastUserScrollActionIndex();
                                    }
                                }
                            }
                            
                        }
                        
                        // Re-Adjust the scroll pane if required
                        if (!userAction) {
                            e.getAdjustable().setValue(value);
                        }
                    }
                });
        
            this.addBody(logContainer);
        } else {
            story.setText(storyText);
        }
    }
    
    /**
     * Registers a new user triggered scroll action related to the display panel.
     * 
     * @param index         The new index of the scroll event.
     */
    public final void updateUserScrollAction(final int index) {
        lastUserScrollTime = System.currentTimeMillis();
        lastUserScrollIndex = index;
    }
    
    /**
     * Returns the last time index the user triggered a scroll event related to the display panel. 
     * 
     * @return              The time index of the last scroll event as <i>UNIX timestamp</i>.
     */
    public final long getLastUserScrollActionTime() {
        return this.lastUserScrollTime;
    }
    
    /**
     * Returns the last scroll index the user positioned the display panel.
     * 
     * @return              The last scroll index the user positioned the display panel as <i>int</i>.
     */
    public final int getLastUserScrollActionIndex() {
        return this.lastUserScrollIndex;
    }
    
    /**
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     * 
     * @param e             An {@link ActionEvent}.
     */
    public final void actionPerformed(final ActionEvent e) {
        // Menu Items
        if (JMenuItem.class.isAssignableFrom(e.getSource().getClass())) {
            String menuItemType = ((JMenuItem) e.getSource()).getName();
            if ("about".equals(menuItemType)) {
                infoDialog(
                        "Version: " + MainConfig.get("REACHABILITY_UI_VERSION ")
                        + "\nAuthor: " + ProjectInfo.REACHABILITY_AGENT_AUTHOR
                );
            }
        }
        
        // Repaint timer
        if (e.getSource().equals(this.getTimer())) {
            this.refresh(true);
        }
    }
    
    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        // Shut down everything...
    }
    
    /**
     * Builds the user interface. The <i>refresh</i> parameter is used for positioning reasons.
     * 
     * @param refresh       Indicates if this method was called with the intention to re-render the user interface.
     */
    public final void refresh(final boolean refresh) {
        int rows = 1;
        
        SortedMap<SensorDataFeederItem, String> news = new TreeMap<SensorDataFeederItem, String>();
        if (this.feeder != null) {
            news = this.feeder.getLog();
        }
        int lines    = news.size() + 1;
        int maxLines = getConfigValueAsInteger("SHOW_MAX_ROWS");
        if (lines > maxLines) {
            lines = maxLines;
        }
        
        // Adapt window size
        int newHeight = getConfigValueAsInteger("DEFAULT_WINDOW_MENU_HEIGHT")
                        + (getConfigValueAsInteger("DEFAULT_ROW_HEIGHT") * lines);
        setWindowHeight(newHeight);
        fixBounds(refresh);
        
        // Define layout
        if (story == null) {
            getBody().setLayout(getGrid(rows));
            getContentPane().add(getBody());
            getBody().removeAll();
        }
        
        // Set story rows
        addStoryRow(news);
            
        // Reload content
        fixBounds(refresh);
        getBody().revalidate();
    }
}
